@@ -1,6 +1,6 @@
 package ClearCase::Wrapper::MGi;
 
-$VERSION = '0.18';
+$VERSION = '0.19';
 
 use warnings;
 use strict;
@@ -431,7 +431,7 @@ sub _Preemptcmt { #return the comments apart: e.g. mklbtype needs discrimination
   if ($ncf or $cf) {
     if ($ncf) {
       $cmd->opts(grep !/^-nc(?:omment)?$/,@opts);
-      $ret = &$fn($cmd, qw(-nc));
+      $ret = $fn->($cmd, qw(-nc));
     } else {
       my $skip = 0;
       for (@opts) {
@@ -448,7 +448,7 @@ sub _Preemptcmt { #return the comments apart: e.g. mklbtype needs discrimination
 	}
       }
       $cmd->opts(@mod);
-      $ret = &$fn($cmd, @copt);
+      $ret = $fn->($cmd, @copt);
     }
   } elsif ($cqf) {
     my $cqe = grep /^-cqe/, @opts;
@@ -460,12 +460,12 @@ sub _Preemptcmt { #return the comments apart: e.g. mklbtype needs discrimination
       if ($cqe) {
 	my $arg = shift @arg;
 	$go = scalar @arg;
-	next if $tst and !&$tst($cmd, $arg);
+	next if $tst and !$tst->($cmd, $arg);
 	$cmd->args($arg);
 	print qq(Comments for "$arg":\n);
       } else {
 	$go = 0;
-	last if $tst and !&$tst($cmd, @arg); #None checked out
+	last if $tst and !$tst->($cmd, @arg); #None checked out
 	print "Comment for all listed objects:\n";
       }
       my $cmt = '';
@@ -478,9 +478,9 @@ sub _Preemptcmt { #return the comments apart: e.g. mklbtype needs discrimination
 	my ($fh, $cfile) = tempfile();
 	print $fh $cmt;
 	close $fh;
-	$ret |= &$fn($cmd, '-cfile', $cfile);
+	$ret |= $fn->($cmd, '-cfile', $cfile);
       } else {
-	$ret |= &$fn($cmd, '-c', $cmt);
+	$ret |= $fn->($cmd, '-c', $cmt);
       }
     }
   }
@@ -489,20 +489,19 @@ sub _Preemptcmt { #return the comments apart: e.g. mklbtype needs discrimination
 sub _Unco {
   use warnings;
   use strict;
-  my $unco = shift;
+  my ($unco, $rc) = (shift, 0);
   $ct = ClearCase::Argv->new({autochomp=>1});
-  my @b0 = grep { m%[\\/]CHECKEDOUT$% }
-    $ct->argv(qw(ls -s -d), $unco->args)->qx;
-  my $rc = $unco->system;
-  map { s%^(.*)[\\/]CHECKEDOUT$%$1% } @b0;
-  my @rm = ();
-  for my $d (@b0) {
-    opendir BR, $d or next;
-    my @f = grep !/^\.\.?$/, readdir BR;
-    closedir BR;
-    push @rm, $d if (scalar @f == 2) and $f[0] eq '0' and $f[1] eq 'LATEST';
+  for my $arg ($unco->args) { # Already sorted if several
+    my $b0 = $ct->argv(qw(ls -s -d), $arg)->qx;
+    $rc |= $unco->args($arg)->system;
+    if ($b0 =~ s%^(.*)[\\/]CHECKEDOUT$%$1%) {
+      opendir BR, $b0 or next;
+      my @f = grep !/^\.\.?$/, readdir BR;
+      closedir BR;
+      $ct->argv(qw(rmbranch -f), $b0)->system
+	if (scalar @f == 2) and $f[0] eq '0' and $f[1] eq 'LATEST';
+    }
   }
-  $ct->argv(qw(rmbranch -f), @rm)->system if @rm;
   return $rc;
 }
 sub _Yesno {
@@ -514,10 +513,13 @@ sub _Yesno {
   for my $arg ($cmd->args) {
     $cmd->opts(@opts); #reset
     if ($test) {
-      my $res = $test->args($arg)->qx;
-      if (!$res or $res =~ /^cleartool: Error:/) {
+      my $res = $test->($arg);
+      if (!$res) {
 	warn ($res or Msg('E', $errmsg . '"' . $arg . "\".\n"));
 	next;
+      } elsif ($res == -1) { # Skip interactive part
+	$ret |= $fn->($cmd);
+	next;
       }
     }
     printf $yn->{format}, $arg;
@@ -531,7 +533,7 @@ sub _Yesno {
     if ($yn->{opt}->{$ans}) {
       $cmd->opts(@opts, $yn->{opt}->{$ans});
       $cmd->args($arg);
-      $ret |= &{$fn}($cmd);
+      $ret |= $fn->($cmd);
     } else {
       $ret = 1;
     }
@@ -554,9 +556,10 @@ sub _Checkin {
     );
     my $lsco = ClearCase::Argv->lsco([qw(-cview -s)]);
     $lsco->stderr(1);
+    my $ok = sub { return $lsco->args(shift)->qx? 1:0 };
     my $err = 'Unable to find checked out version for ';
     my $run = sub { my $ci = shift; $ci->system };
-    _Yesno($ci, $run, \%yn, $lsco, $err); #only one arg: may exit
+    _Yesno($ci, $run, \%yn, $ok, $err); #only one arg: may exit
   } else {
     return $ci->system;
   }
@@ -564,7 +567,7 @@ sub _Checkin {
 
 =head1 NAME
 
-ClearCase::Wrapper::MGi - Marc Girod's contributed cleartool wrapper functions
+ClearCase::Wrapper::MGi - Support for an alternative to UCM.
 
 =head1 SYNOPSIS
 
@@ -572,6 +575,9 @@ This is an C<overlay module> for B<ClearCase::Wrapper> containing Marc
 Girod's non-standard extensions. See C<perldoc ClearCase::Wrapper> (by
 David Boyce) for more details.
 
+The alternative to UCM consists in a novel branching model, and a concept
+of incremental types.
+
 =head1 CLEARTOOL EXTENSIONS
 
 =over 12
@@ -746,9 +752,19 @@ sub mkbranch {
   $mkbranch->parse(qw(nwarn nco ptime cquery|cqeach nc c|cfile=s));
   my @args = $mkbranch->args;
   my $bt = shift @args;
-  $bt =~ s/^brtype:(.*)$/$1/;
-  die if $ct->argv(qw(des -s), "brtype:$bt")->stdout(0)->system;
   map { $_ = glob($_) } @args if MSWIN;
+  $bt =~ s/^brtype:(.*)$/$1/;
+  my %v;
+  if ($bt =~ /\@/) {
+    $v{''} = 1;
+  } else {
+    for my $a (@args) {
+      $v{$ct->argv(qw(des -fmt), '@%n', "vob:$a")->stderr(0)->qx}++;
+    }
+  }
+  for (keys %v) {
+    die if $ct->argv(qw(des -s), "brtype:$bt$_")->stdout(0)->system;
+  }
   $mkbranch->args(@args);
   $mkbranch->{bt} = $bt;
   $mkbranch->{ver} = $ver;
@@ -762,7 +778,7 @@ an equal basis as parents on the same physical branch.
 In case there are multiple parents, consider the one on the same branch as
 'more equal than the others' (least surprise principle).
 
-Preserve the (Wrapper default) assumption of a B<-pred> flag, is only one
+Preserve the (Wrapper default) assumption of a B<-pred> flag, if only one
 argument is given.
 
 =cut
@@ -772,6 +788,12 @@ sub diff {
     $_ = readlink if -l && defined readlink;
   }
   push(@ARGV, qw(-dir)) if @ARGV == 1;
+  my $limit = 0;
+  if (my @num = grep /^-\d+$/, @ARGV) {
+    @ARGV = grep !/^-\d+$/, @ARGV;
+    die Msg('E', "incompatible flags: @num") if @num > 1;
+    $limit = -int($num[0]);
+  }
   my $diff = ClearCase::Argv->new(@ARGV);
   $diff->autochomp(1);
   $diff->ipc(1) unless $diff->ctcmd(1);
@@ -786,6 +808,7 @@ sub diff {
   $diff->opts(grep !/-pred/, @opts) if $pred;
   $diff->args(@elems);
   $ct = $diff->clone();
+  my $rc = 0;
   for my $e (@elems) {
     my ($ele, $ver, $type) =
       $ct->argv(qw(des -fmt), '%En\n%En@@%Vn\n%m', $e)->qx;
@@ -798,11 +821,17 @@ sub diff {
     my $bra = $1 if $ver =~ m%^(.*?)/(?:\d+|CHECKEDOUT)$%;
     my %gen = _Parsevtree($ele, 1, $ver);
     my $p = $gen{$ver}{'parents'};
+    $p = $gen{$p->[0]}{'parents'} while $p and $limit--;
+    if (!$p) {
+      warn Msg('E', "No predecessor version to compare to: $e");
+      $rc = 1;
+      next;
+    }
     my ($brp) = grep { m%^$bra/\d+$% } @{$p};
     $ver = $ele if $ver =~ m%/CHECKEDOUT$%;
-    $diff->args($brp? $brp : $p->[0], $ver)->system;
+    $rc |= $diff->args($brp? $brp : $p->[0], $ver)->system;
   }
-  exit $?;
+  exit $rc;
 }
 
 =item * UNCHECKOUT
@@ -838,8 +867,15 @@ sub uncheckout {
     );
     my $lsco = ClearCase::Argv->lsco([qw(-cview -s)]);
     $lsco->stderr(1);
+    my $dm = ClearCase::Argv->des([qw(-fmt %m)]);
+    $dm->stderr(1);
+    my $ok = sub {
+      my $e = shift;
+      return -1 if $dm->args($e)->qx =~ /^directory/;
+      return $lsco->args($e)->qx? 1:0;
+    };
     my $err = 'Unable to find checked out version for ';
-    _Yesno($unco, \&_Unco, \%yn, $lsco, $err);
+    _Yesno($unco, \&_Unco, \%yn, $ok, $err);
   }
 }
 
@@ -955,11 +991,15 @@ sub _GenMkTypeSub {
       }
     } elsif (%opt) {
       map { s/^${type}:(.*)$/$1/ } @args;
-      my @a = @args;
-      my @vob = grep { s/.*\@(.*)$/$1/ } @a;
+      my (@a, $unkvob) = @args;
+      my @vob = grep { s/[^@]+\@(.*)$/$1/ } @a;
+      for (@vob) {
+	$unkvob++ if $silent->argv(qw(des -s), "vob:$_")->system;
+      }
+      return 1 if $unkvob;
       push @vob, $ct->argv(qw(des -s vob:.))->stderr(0)->qx
 	if grep !/@/, @args;
-      die  Msg('E', qq(Unable to determine VOB for pathname ".".\n))
+      die  Msg('E', qq(Unable to determine VOB for pathname ".".))
 	unless @vob;
       _Ensuretypes(@vob);
       if ($rep) {
@@ -1054,6 +1094,9 @@ sub _GenMkTypeSub {
 	    }
 	  } keys %pair;
 	} elsif ($opt{increment}) { # increment
+	  $ntype->opts(@cmt, $ntype->opts);
+	  my $lct = ClearCase::Argv->new(); #Not autochomp
+	  my ($fl, $loaded) = $ENV{FORCELOCK};
 	  INCT: for my $t (@args) {
 	    my ($pt, $lck) = "$type:$t";
 	    if (!$ct->argv(qw(des -s), $pt)->stderr(0)->qx) {
@@ -1062,10 +1105,11 @@ sub _GenMkTypeSub {
 	    }
 	    my ($pair) = grep s/^\s*(.*) -> $type:(.*)\@(.*)$/$1,$2,$3/,
 	      $ct->argv(qw(des -l -ahl), $eqhl, $pt)->stderr(0)->qx;
-	    my ($hlk, $prev, $vob) = split ',', $pair if $pair;
-	    next INCT unless $prev;
-	    my $lct = ClearCase::Argv->new(); #Not autochomp
-	    my ($fl, $loaded) = $ENV{FORCELOCK};
+	    my ($hlk, $prev, $vob) = split /,/, $pair if $pair;
+	    if (!$prev) {
+	      warn Msg('E', "Not a family type: $t");
+	      next INCT;
+	    }
 	    my ($t1) = $t =~ /^(.*?)(@|$)/;
 	    for my $l ($t1, $prev) {
 	      if ($ct->argv(qw(lslock -s), "lbtype:$l\@$vob")->stderr(0)->qx) {
@@ -1092,12 +1136,14 @@ sub _GenMkTypeSub {
 	      }
 	    }
 	    if ($prev =~ /^(.*)_(\d+)(?:\.(\d+))?$/) {
-	      my ($base, $maj, $min) = ($1, $2, $3);
-	      my $new = "${base}_" .
-		(defined($min)? $maj . '.' . ++$min : ++$maj);
-	      my $p1 = $prev;
-	      map { $_ .= $1 } ($new, $p1) if $t =~ /^.*(@.*)$/;
-	      $ntype->opts(@cmt, $ntype->opts);
+	      my ($base, $maj, $min, $new) = ($1, $2, $3);
+	      my $ext = ($t =~ /^.*(@.*)$/)? $1 : '';
+	      my $p1 = $prev . $ext;
+	      do {
+		$new = "${base}_" .
+		  (defined($min)? $maj . '.' . ++$min : ++$maj);
+		$new .= $ext;
+	      } while ($ct->argv(qw(des -s), "$type:$new")->stderr(0)->qx);
 	      $ntype->args($new)->system and exit 1;
 	      $silent->argv('rmhlink', $hlk)->system;
 	      $silent->argv(qw(mkhlink -nc), $eqhl,
@@ -1182,6 +1228,12 @@ sub mklbtype {
   $ntype->parse(qw(global|ordinary vpelement|vpbranch|vpversion
 		   pbranch|shared gt|ge|lt|le|enum|default|vtype=s
 		   cquery|cqeach nc c|cfile=s));
+  if (!$ntype->args) {
+    warn Msg('E', 'Type name required.');
+    @ARGV = qw(help mklbtype);
+    ClearCase::Wrapper->help();
+    return 1;
+  }
   $ntype->{fopts} = \%opt;
   $ntype->{rep} = $opt{archive}? 1 : $rep;
   my $tst = $ntype->{rep}? _GenExTypeSub('lbtype') : 0;
@@ -1293,7 +1345,7 @@ sub lock {
     $lock->opts($lock->opts, '-nusers', join(',', sort keys %nusers))
       if %nusers;
   } elsif (($nusers or $opt{allow}) and
-	     (!$currlock or $opt{iflocked}) or $lock->flag('replace')) {
+	     (!$currlock or $opt{iflocked} or $lock->flag('replace'))) {
     $lock->opts($lock->opts, '-nusers', ($nusers or $opt{allow}));
   }
   if ($currlock and !$lock->flag('replace')) {
@@ -1655,6 +1707,12 @@ sub checkin {
   }
   # Now do auto-aggregation on the remaining args.
   my @elems = AutoCheckedOut($opt{ok}, $ci->args); # may exit
+  if (!@elems) {
+    warn Msg('E', 'Element pathname required.');
+    @ARGV = qw(help checkin);
+    ClearCase::Wrapper->help();
+    return 1;
+  }
   # Turn symbolic links into their targets so CC will "do the right thing".
   for (@elems) {
     $_ = readlink if -l && defined readlink;
@@ -1704,8 +1762,10 @@ sub rmbranch {
       instruct => "Please answer with one of the following: yes, no\n",
       opt      => \%forceorabort,
     );
-    my $exists = ClearCase::Argv->des([qw(-s)]);
-    $exists->stderr(1);
+    my $des = ClearCase::Argv->des([qw(-s)]);
+    $des->stderr(0);
+    $des->stdout(0);
+    my $exists = sub { return $des->args(shift)->system? 0:1 };
     my $err = q(Pathname not found: );
     my $run = sub { my $rmbranch = shift; $rmbranch->system };
     _Yesno($rmbranch, $run, \%yn, $exists, $err);
@@ -1717,7 +1777,7 @@ sub rmbranch {
 =head1 COPYRIGHT AND LICENSE
 
 Copyright (c) 2007 IONA Technologies PLC (until v0.05),
-2008-2009 Marc Girod (marc.girod@gmail.com) for later versions.
+2008-2010 Marc Girod (marc.girod@gmail.com) for later versions.
 All rights reserved.
 This Perl program is free software; you may redistribute it
 and/or modify it under the same terms as Perl itself.
